@@ -3,13 +3,13 @@ from __future__ import annotations
 import os
 import re
 import sqlite3
+import subprocess
 from collections import defaultdict
 from pathlib import Path
 from typing import TYPE_CHECKING
 
 import dill as pickle
 from junitparser.xunit2 import JUnitXml
-import subprocess
 from lxml.etree import XMLParser, parse
 
 from codeflash.cli_cmds.console import DEBUG_MODE, console, logger
@@ -20,13 +20,13 @@ from codeflash.code_utils.code_utils import (
     module_name_from_file_path,
 )
 from codeflash.discovery.discover_unit_tests import discover_parameters_unittest
-from codeflash.models.models import CoverageData, TestFiles
-from codeflash.verification.test_results import FunctionTestInvocation, InvocationId, TestResults
+from codeflash.models.models import FunctionTestInvocation, InvocationId, TestResults, TestType, VerificationType
+from codeflash.verification.coverage_utils import CoverageUtils
 
 if TYPE_CHECKING:
     import subprocess
 
-    from codeflash.models.models import CodeOptimizationContext
+    from codeflash.models.models import CodeOptimizationContext, CoverageData, TestFiles
     from codeflash.verification.verification_utils import TestConfig
 
 
@@ -36,6 +36,10 @@ def parse_func(file_path: Path) -> XMLParser:
     return parse(file_path, xml_parser)
 
 
+matches_re_start = re.compile(r"!\$######(.*?):(.*?)([^\.:]*?):(.*?):(.*?):(.*?)######\$!\n")
+matches_re_end = re.compile(r"!######(.*?):(.*?)([^\.:]*?):(.*?):(.*?):(.*?)######!")
+
+
 def parse_test_return_values_bin(file_location: Path, test_files: TestFiles, test_config: TestConfig) -> TestResults:
     test_results = TestResults()
     if not file_location.exists():
@@ -44,56 +48,56 @@ def parse_test_return_values_bin(file_location: Path, test_files: TestFiles, tes
         return test_results
 
     with file_location.open("rb") as file:
-        while file:
-            len_next_bytes = file.read(4)
-            if not len_next_bytes:
-                return test_results
-            len_next = int.from_bytes(len_next_bytes, byteorder="big")
-            encoded_test_name = file.read(len_next).decode("ascii")
-            duration_bytes = file.read(8)
-            duration = int.from_bytes(duration_bytes, byteorder="big")
-            len_next_bytes = file.read(4)
-            if not len_next_bytes:
-                return test_results
-            len_next = int.from_bytes(len_next_bytes, byteorder="big")
-            try:
+        try:
+            while file:
+                len_next_bytes = file.read(4)
+                if not len_next_bytes:
+                    return test_results
+                len_next = int.from_bytes(len_next_bytes, byteorder="big")
+                encoded_test_bytes = file.read(len_next)
+                encoded_test_name = encoded_test_bytes.decode("ascii")
+                duration_bytes = file.read(8)
+                duration = int.from_bytes(duration_bytes, byteorder="big")
+                len_next_bytes = file.read(4)
+                len_next = int.from_bytes(len_next_bytes, byteorder="big")
                 test_pickle_bin = file.read(len_next)
-            except Exception as e:
-                if DEBUG_MODE:
-                    logger.exception(f"Failed to load pickle file. Exception: {e}")
-                return test_results
-            loop_index_bytes = file.read(8)
-            loop_index = int.from_bytes(loop_index_bytes, byteorder="big")
-            len_next_bytes = file.read(4)
-            len_next = int.from_bytes(len_next_bytes, byteorder="big")
-            invocation_id = file.read(len_next).decode("ascii")
+                loop_index_bytes = file.read(8)
+                loop_index = int.from_bytes(loop_index_bytes, byteorder="big")
+                len_next_bytes = file.read(4)
+                len_next = int.from_bytes(len_next_bytes, byteorder="big")
+                invocation_id_bytes = file.read(len_next)
+                invocation_id = invocation_id_bytes.decode("ascii")
 
-            invocation_id_object = InvocationId.from_str_id(encoded_test_name, invocation_id)
-            test_file_path = file_path_from_module_name(
-                invocation_id_object.test_module_path, test_config.tests_project_rootdir
-            )
-
-            test_type = test_files.get_test_type_by_instrumented_file_path(test_file_path)
-            try:
-                test_pickle = pickle.loads(test_pickle_bin) if loop_index == 1 else None
-            except Exception as e:
-                if DEBUG_MODE:
-                    logger.exception(f"Failed to load pickle file for {encoded_test_name} Exception: {e}")
-                continue
-            assert test_type is not None, f"Test type not found for {test_file_path}"
-            test_results.add(
-                function_test_invocation=FunctionTestInvocation(
-                    loop_index=loop_index,
-                    id=invocation_id_object,
-                    file_name=test_file_path,
-                    did_pass=True,
-                    runtime=duration,
-                    test_framework=test_config.test_framework,
-                    test_type=test_type,
-                    return_value=test_pickle,
-                    timed_out=False,
+                invocation_id_object = InvocationId.from_str_id(encoded_test_name, invocation_id)
+                test_file_path = file_path_from_module_name(
+                    invocation_id_object.test_module_path, test_config.tests_project_rootdir
                 )
-            )
+
+                test_type = test_files.get_test_type_by_instrumented_file_path(test_file_path)
+                try:
+                    test_pickle = pickle.loads(test_pickle_bin) if loop_index == 1 else None
+                except Exception as e:
+                    if DEBUG_MODE:
+                        logger.exception(f"Failed to load pickle file for {encoded_test_name} Exception: {e}")
+                    continue
+                assert test_type is not None, f"Test type not found for {test_file_path}"
+                test_results.add(
+                    function_test_invocation=FunctionTestInvocation(
+                        loop_index=loop_index,
+                        id=invocation_id_object,
+                        file_name=test_file_path,
+                        did_pass=True,
+                        runtime=duration,
+                        test_framework=test_config.test_framework,
+                        test_type=test_type,
+                        return_value=test_pickle,
+                        timed_out=False,
+                        verification_type=VerificationType.FUNCTION_CALL,
+                    )
+                )
+        except Exception as e:
+            logger.warning(f"Failed to parse test results from {file_location}. Exception: {e}")
+            return test_results
     return test_results
 
 
@@ -103,43 +107,59 @@ def parse_sqlite_test_results(sqlite_file_path: Path, test_files: TestFiles, tes
         logger.warning(f"No test results for {sqlite_file_path} found.")
         console.rule()
         return test_results
+    db = None
     try:
         db = sqlite3.connect(sqlite_file_path)
         cur = db.cursor()
         data = cur.execute(
             "SELECT test_module_path, test_class_name, test_function_name, "
-            "function_getting_tested, loop_index, iteration_id, runtime, return_value FROM test_results"
+            "function_getting_tested, loop_index, iteration_id, runtime, return_value,verification_type FROM test_results"
         ).fetchall()
+    except Exception as e:
+        logger.warning(f"Failed to parse test results from {sqlite_file_path}. Exception: {e}")
+        if db is not None:
+            db.close()
+        return test_results
     finally:
         db.close()
     for val in data:
         try:
             test_module_path = val[0]
+            test_class_name = val[1] if val[1] else None
+            test_function_name = val[2] if val[2] else None
+            function_getting_tested = val[3]
             test_file_path = file_path_from_module_name(test_module_path, test_config.tests_project_rootdir)
-            # TODO : this is because sqlite writes original file module path. Should make it consistent
-            test_type = test_files.get_test_type_by_original_file_path(test_file_path)
             loop_index = val[4]
+            iteration_id = val[5]
+            runtime = val[6]
+            verification_type = val[8]
+            if verification_type in {VerificationType.INIT_STATE_FTO, VerificationType.INIT_STATE_HELPER}:
+                test_type = TestType.INIT_STATE_TEST
+            else:
+                # TODO : this is because sqlite writes original file module path. Should make it consistent
+                test_type = test_files.get_test_type_by_original_file_path(test_file_path)
             try:
                 ret_val = (pickle.loads(val[7]) if loop_index == 1 else None,)
-            except Exception:
+            except Exception:  # noqa: S112
                 continue
             test_results.add(
                 function_test_invocation=FunctionTestInvocation(
                     loop_index=loop_index,
                     id=InvocationId(
-                        test_module_path=val[0],
-                        test_class_name=val[1],
-                        test_function_name=val[2],
-                        function_getting_tested=val[3],
-                        iteration_id=val[5],
+                        test_module_path=test_module_path,
+                        test_class_name=test_class_name,
+                        test_function_name=test_function_name,
+                        function_getting_tested=function_getting_tested,
+                        iteration_id=iteration_id,
                     ),
                     file_name=test_file_path,
                     did_pass=True,
-                    runtime=val[6],
+                    runtime=runtime,
                     test_framework=test_config.test_framework,
                     test_type=test_type,
                     return_value=ret_val,
                     timed_out=False,
+                    verification_type=VerificationType(verification_type) if verification_type else None,
                 )
             )
         except Exception:
@@ -173,7 +193,7 @@ def parse_test_xml(
     for suite in xml:
         for testcase in suite:
             class_name = testcase.classname
-            test_file_name = suite._elem.attrib.get("file")
+            test_file_name = suite._elem.attrib.get("file")  # noqa: SLF001
             if (
                 test_file_name == f"unittest{os.sep}loader.py"
                 and class_name == "unittest.loader._FailedTest"
@@ -243,8 +263,18 @@ def parse_test_xml(
                     message = testcase.result[0].message.lower()
                     if "timed out" in message:
                         timed_out = True
-            matches = re.findall(r"!######(.*?):(.*?)([^\.:]*?):(.*?):(.*?):(.*?)######!", testcase.system_out or "")
-            if not matches or not len(matches):
+
+            sys_stdout = testcase.system_out or ""
+            begin_matches = list(matches_re_start.finditer(sys_stdout))
+            end_matches = {}
+            for match in matches_re_end.finditer(sys_stdout):
+                groups = match.groups()
+                if len(groups[5].split(":")) > 1:
+                    iteration_id = groups[5].split(":")[0]
+                    groups = groups[:5] + (iteration_id,)
+                end_matches[groups] = match
+
+            if not begin_matches or not begin_matches:
                 test_results.add(
                     FunctionTestInvocation(
                         loop_index=loop_index,
@@ -252,7 +282,7 @@ def parse_test_xml(
                             test_module_path=test_module_path,
                             test_class_name=test_class,
                             test_function_name=test_function,
-                            function_getting_tested="",  # FIXME
+                            function_getting_tested="",  # TODO: Fix this
                             iteration_id="",
                         ),
                         file_name=test_file_path,
@@ -262,28 +292,46 @@ def parse_test_xml(
                         test_type=test_type,
                         return_value=None,
                         timed_out=timed_out,
+                        stdout="",
                     )
                 )
 
             else:
-                for match in matches:
+                for match_index, match in enumerate(begin_matches):
+                    groups = match.groups()
+                    end_match = end_matches.get(groups)
+                    iteration_id, runtime = groups[5], None
+                    if end_match:
+                        stdout = sys_stdout[match.end() : end_match.start()]
+                        split_val = end_match.groups()[5].split(":")
+                        if len(split_val) > 1:
+                            iteration_id = split_val[0]
+                            runtime = int(split_val[1])
+                        else:
+                            iteration_id, runtime = split_val[0], None
+                    elif match_index == len(begin_matches) - 1:
+                        stdout = sys_stdout[match.end() :]
+                    else:
+                        stdout = sys_stdout[match.end() : begin_matches[match_index + 1].start()]
+
                     test_results.add(
                         FunctionTestInvocation(
-                            loop_index=int(match[4]),
+                            loop_index=int(groups[4]),
                             id=InvocationId(
-                                test_module_path=match[0],
-                                test_class_name=None if match[1] == "" else match[1][:-1],
-                                test_function_name=match[2],
-                                function_getting_tested=match[3],
-                                iteration_id=match[5],
+                                test_module_path=groups[0],
+                                test_class_name=None if groups[1] == "" else groups[1][:-1],
+                                test_function_name=groups[2],
+                                function_getting_tested=groups[3],
+                                iteration_id=iteration_id,
                             ),
                             file_name=test_file_path,
-                            runtime=None,
+                            runtime=runtime,
                             test_framework=test_config.test_framework,
                             did_pass=result,
                             test_type=test_type,
                             return_value=None,
                             timed_out=timed_out,
+                            stdout=stdout,
                         )
                     )
 
@@ -325,22 +373,22 @@ def merge_test_results(
                 test_function_name = new_test_function_name
 
         grouped_xml_results[
-            result.id.test_module_path
+            (result.id.test_module_path or "")
             + ":"
             + (result.id.test_class_name or "")
             + ":"
-            + test_function_name
+            + (test_function_name or "")
             + ":"
             + str(result.loop_index)
         ].add(result)
 
     for result in bin_test_results:
         grouped_bin_results[
-            result.id.test_module_path
+            (result.id.test_module_path or "")
             + ":"
             + (result.id.test_class_name or "")
             + ":"
-            + result.id.test_function_name
+            + (result.id.test_function_name or "")
             + ":"
             + str(result.loop_index)
         ].add(result)
@@ -368,6 +416,10 @@ def merge_test_results(
                         test_type=xml_result.test_type,
                         return_value=result_bin.return_value,
                         timed_out=xml_result.timed_out,
+                        verification_type=VerificationType(result_bin.verification_type)
+                        if result_bin.verification_type
+                        else None,
+                        stdout=xml_result.stdout,
                     )
                 )
         elif xml_results.test_results[0].id.iteration_id is not None:
@@ -394,6 +446,10 @@ def merge_test_results(
                         timed_out=xml_result.timed_out
                         if bin_result.runtime is None
                         else False,  # If runtime was measured in the bin file, then the testcase did not time out
+                        verification_type=VerificationType(bin_result.verification_type)
+                        if bin_result.verification_type
+                        else None,
+                        stdout=xml_result.stdout,
                     )
                 )
         else:
@@ -417,6 +473,10 @@ def merge_test_results(
                         test_type=bin_result.test_type,
                         return_value=bin_result.return_value,
                         timed_out=xml_result.timed_out,  # only the xml gets the timed_out flag
+                        verification_type=VerificationType(bin_result.verification_type)
+                        if bin_result.verification_type
+                        else None,
+                        stdout=xml_result.stdout,
                     )
                 )
 
@@ -430,7 +490,8 @@ def parse_test_results(
     optimization_iteration: int,
     function_name: str | None,
     source_file: Path | None,
-    coverage_file: Path | None,
+    coverage_database_file: Path | None,
+    coverage_config_file: Path | None,
     code_context: CodeOptimizationContext | None = None,
     run_result: subprocess.CompletedProcess | None = None,
     unittest_loop_index: int | None = None,
@@ -466,19 +527,20 @@ def parse_test_results(
 
     get_run_tmp_file(Path(f"test_return_values_{optimization_iteration}.bin")).unlink(missing_ok=True)
 
+    get_run_tmp_file(Path("pytest_results.xml")).unlink(missing_ok=True)
+    get_run_tmp_file(Path("unittest_results.xml")).unlink(missing_ok=True)
     get_run_tmp_file(Path(f"test_return_values_{optimization_iteration}.sqlite")).unlink(missing_ok=True)
     results = merge_test_results(test_results_xml, test_results_bin_file, test_config.test_framework)
 
     all_args = False
-    if coverage_file and source_file and code_context and function_name:
+    if coverage_database_file and source_file and code_context and function_name:
         all_args = True
-        coverage = CoverageData.load_from_coverage_file(
-            coverage_file_path=coverage_file,
+        coverage = CoverageUtils.load_from_sqlite_database(
+            database_path=coverage_database_file,
+            config_path=coverage_config_file,
             source_code_path=source_file,
             code_context=code_context,
             function_name=function_name,
         )
-        coverage_file.unlink(missing_ok=True)
-        Path(".coverage").unlink(missing_ok=True)
         coverage.log_coverage()
     return results, coverage if all_args else None
